"""Constants for the Deluge integration."""
import logging

CONF_WEB_PORT = "web_port"
DEFAULT_NAME = "Deluge"
DEFAULT_RPC_PORT = 58846
DEFAULT_WEB_PORT = 8112
DHT_UPLOAD = 1000
DHT_DOWNLOAD = 1000
DOMAIN = "deluge"

LOGGER = logging.getLogger(__package__)
