"""Constants for the Uptime integration."""
from typing import Final

from homeassistant.const import Platform

DOMAIN: Final = "uptime"
PLATFORMS: Final = [Platform.SENSOR]

DEFAULT_NAME: Final = "Uptime"
