"""Automatically generated by hassfest.

To update, run python3 -m script.hassfest
"""

# fmt: off

FLOWS = {
    "integration": [
        "abode",
        "accuweather",
        "acmeda",
        "adax",
        "adguard",
        "advantage_air",
        "aemet",
        "agent_dvr",
        "airly",
        "airnow",
        "airthings",
        "airtouch4",
        "airvisual",
        "airzone",
        "alarmdecoder",
        "almond",
        "ambee",
        "amberelectric",
        "ambiclimate",
        "ambient_station",
        "androidtv",
        "apple_tv",
        "arcam_fmj",
        "aseko_pool_live",
        "asuswrt",
        "atag",
        "august",
        "aurora",
        "aurora_abb_powerone",
        "aussie_broadband",
        "awair",
        "axis",
        "azure_devops",
        "azure_event_hub",
        "balboa",
        "blebox",
        "blink",
        "bmw_connected_drive",
        "bond",
        "bosch_shc",
        "braviatv",
        "broadlink",
        "brother",
        "brunt",
        "bsblan",
        "buienradar",
        "canary",
        "cast",
        "cert_expiry",
        "cloudflare",
        "co2signal",
        "coinbase",
        "control4",
        "coolmaster",
        "coronavirus",
        "cpuspeed",
        "crownstone",
        "daikin",
        "deconz",
        "deluge",
        "denonavr",
        "devolo_home_control",
        "devolo_home_network",
        "dexcom",
        "dialogflow",
        "directv",
        "discord",
        "dlna_dmr",
        "dlna_dms",
        "dnsip",
        "doorbird",
        "dsmr",
        "dunehd",
        "dynalite",
        "eafm",
        "ecobee",
        "econet",
        "efergy",
        "elgato",
        "elkm1",
        "elmax",
        "emonitor",
        "emulated_roku",
        "enocean",
        "enphase_envoy",
        "environment_canada",
        "epson",
        "esphome",
        "evil_genius_labs",
        "ezviz",
        "faa_delays",
        "fibaro",
        "filesize",
        "fireservicerota",
        "fivem",
        "fjaraskupan",
        "flick_electric",
        "flipr",
        "flo",
        "flume",
        "flunearyou",
        "flux_led",
        "forecast_solar",
        "forked_daapd",
        "foscam",
        "freebox",
        "freedompro",
        "fritz",
        "fritzbox",
        "fritzbox_callmonitor",
        "fronius",
        "garages_amsterdam",
        "gdacs",
        "generic",
        "geofency",
        "geonetnz_quakes",
        "geonetnz_volcano",
        "gios",
        "github",
        "glances",
        "goalzero",
        "gogogate2",
        "goodwe",
        "google",
        "google_travel_time",
        "gpslogger",
        "gree",
        "growatt_server",
        "guardian",
        "habitica",
        "hangouts",
        "harmony",
        "heos",
        "hisense_aehw4a1",
        "hive",
        "hlk_sw16",
        "home_connect",
        "home_plus_control",
        "homekit",
        "homekit_controller",
        "homematicip_cloud",
        "homewizard",
        "honeywell",
        "huawei_lte",
        "hue",
        "huisbaasje",
        "hunterdouglas_powerview",
        "hvv_departures",
        "hyperion",
        "ialarm",
        "iaqualink",
        "icloud",
        "ifttt",
        "insteon",
        "intellifire",
        "ios",
        "iotawatt",
        "ipma",
        "ipp",
        "iqvia",
        "islamic_prayer_times",
        "iss",
        "isy994",
        "izone",
        "jellyfin",
        "juicenet",
        "kaleidescape",
        "keenetic_ndms2",
        "kmtronic",
        "knx",
        "kodi",
        "konnected",
        "kostal_plenticore",
        "kraken",
        "kulersky",
        "launch_library",
        "life360",
        "lifx",
        "litejet",
        "litterrobot",
        "local_ip",
        "locative",
        "logi_circle",
        "lookin",
        "luftdaten",
        "lutron_caseta",
        "lyric",
        "mailgun",
        "mazda",
        "meater",
        "melcloud",
        "met",
        "met_eireann",
        "meteo_france",
        "meteoclimatic",
        "metoffice",
        "mikrotik",
        "mill",
        "minecraft_server",
        "mjpeg",
        "mobile_app",
        "modem_callerid",
        "modern_forms",
        "moehlenhoff_alpha2",
        "monoprice",
        "moon",
        "motion_blinds",
        "motioneye",
        "mqtt",
        "mullvad",
        "mutesync",
        "myq",
        "mysensors",
        "nam",
        "nanoleaf",
        "neato",
        "nest",
        "netatmo",
        "netgear",
        "nexia",
        "nfandroidtv",
        "nightscout",
        "nina",
        "nmap_tracker",
        "notion",
        "nuheat",
        "nuki",
        "nut",
        "nws",
        "nzbget",
        "octoprint",
        "omnilogic",
        "oncue",
        "ondilo_ico",
        "onewire",
        "onvif",
        "open_meteo",
        "opengarage",
        "opentherm_gw",
        "openuv",
        "openweathermap",
        "overkiz",
        "ovo_energy",
        "owntracks",
        "p1_monitor",
        "panasonic_viera",
        "peco",
        "philips_js",
        "pi_hole",
        "picnic",
        "plaato",
        "plex",
        "plugwise",
        "plum_lightpad",
        "point",
        "poolsense",
        "powerwall",
        "profiler",
        "progettihwsw",
        "prosegur",
        "ps4",
        "pure_energie",
        "pvoutput",
        "pvpc_hourly_pricing",
        "qnap_qsw",
        "rachio",
        "radio_browser",
        "rainforest_eagle",
        "rainmachine",
        "rdw",
        "recollect_waste",
        "renault",
        "rfxtrx",
        "ridwell",
        "ring",
        "risco",
        "rituals_perfume_genie",
        "roku",
        "roomba",
        "roon",
        "rpi_power",
        "rtsp_to_webrtc",
        "ruckus_unleashed",
        "sabnzbd",
        "samsungtv",
        "screenlogic",
        "season",
        "sense",
        "senseme",
        "sensibo",
        "sentry",
        "senz",
        "sharkiq",
        "shelly",
        "shopping_list",
        "sia",
        "simplisafe",
        "sleepiq",
        "slimproto",
        "sma",
        "smappee",
        "smart_meter_texas",
        "smartthings",
        "smarttub",
        "smhi",
        "sms",
        "solaredge",
        "solarlog",
        "solax",
        "soma",
        "somfy",
        "somfy_mylink",
        "sonarr",
        "songpal",
        "sonos",
        "speedtestdotnet",
        "spider",
        "spotify",
        "sql",
        "squeezebox",
        "srp_energy",
        "starline",
        "steam_online",
        "steamist",
        "stookalert",
        "subaru",
        "sun",
        "surepetcare",
        "switchbot",
        "switcher_kis",
        "syncthing",
        "syncthru",
        "synology_dsm",
        "system_bridge",
        "tado",
        "tailscale",
        "tankerkoenig",
        "tasmota",
        "tautulli",
        "tellduslive",
        "tesla_wall_connector",
        "tibber",
        "tile",
        "tolo",
        "tomorrowio",
        "toon",
        "totalconnect",
        "tplink",
        "traccar",
        "tractive",
        "tradfri",
        "trafikverket_ferry",
        "trafikverket_train",
        "trafikverket_weatherstation",
        "transmission",
        "tuya",
        "twentemilieu",
        "twilio",
        "twinkly",
        "ukraine_alarm",
        "unifi",
        "unifiprotect",
        "upb",
        "upcloud",
        "upnp",
        "uptime",
        "uptimerobot",
        "vallox",
        "velbus",
        "venstar",
        "vera",
        "verisure",
        "version",
        "vesync",
        "vicare",
        "vilfo",
        "vizio",
        "vlc_telnet",
        "volumio",
        "vulcan",
        "wallbox",
        "watttime",
        "waze_travel_time",
        "webostv",
        "wemo",
        "whirlpool",
        "whois",
        "wiffi",
        "wilight",
        "withings",
        "wiz",
        "wled",
        "wolflink",
        "xbox",
        "xiaomi_aqara",
        "xiaomi_miio",
        "yale_smart_alarm",
        "yamaha_musiccast",
        "yeelight",
        "youless",
        "zerproc",
        "zha",
        "zwave_js",
        "zwave_me"
    ],
    "helper": [
        "derivative",
        "group",
        "integration",
        "min_max",
        "switch_as_x",
        "threshold",
        "tod",
        "utility_meter"
    ]
}
