"""Constants for the QNAP QSW integration."""

from typing import Final

ATTR_MAX: Final = "max"

DOMAIN: Final = "qnap_qsw"
MANUFACTURER: Final = "QNAP"

RPM: Final = "rpm"

QSW_TIMEOUT_SEC: Final = 25
